/*
 * Copyright (C) 2006-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file dccp_send.c paraslash's dccp sender */

/*
 * based on server.c of dccp-cs-0.01.tar.bz2,
 * (C) 2005 Ian McDonald <imcdnzl@gmail.com>
 */

#include "server.h"
#include "net.h"
#include "list.h"
#include "vss.h"
#include "send.h"
#include "dccp.h"
#include "error.h"
#include "string.h"
#include "fd.h"
#include "close_on_fork.h"
#include "server.cmdline.h"

/** the list of connected clients **/
static struct list_head clients;
static int listen_fd = -1;
static struct sender *self;

/** describes one connected client */
struct dccp_client {
	/** the dccp socket */
	int fd;
	/** address information about the client */
	struct sockaddr_in addr;
	/** the position of this client in the client list */
	struct list_head node;
	/** non-zero if audio file header has been sent */
	int header_sent;
};

static void dccp_pre_select( int *max_fileno, fd_set *rfds,
		__a_unused fd_set *wfds)
{
	if (listen_fd < 0)
		return;
	FD_SET(listen_fd, rfds);
	*max_fileno = PARA_MAX(*max_fileno, listen_fd);
}

static void dccp_post_select(fd_set *rfds, __a_unused fd_set *wfds)
{
	struct dccp_client *dc;
	int ret;

	if (!FD_ISSET(listen_fd, rfds))
		return;
	dc = para_calloc(sizeof(struct dccp_client));
	ret = para_accept(listen_fd, &dc->addr, sizeof(struct sockaddr_in));
	if (ret < 0) {
		PARA_ERROR_LOG("%s", PARA_STRERROR(-ret));
		return;
	}
	PARA_NOTICE_LOG("connection from %s\n", inet_ntoa(dc->addr.sin_addr));
	dc->fd = ret;
	para_list_add(&dc->node, &clients);
	add_close_on_fork_list(dc->fd);
	mark_fd_nonblock(dc->fd);
}

static int dccp_open(void)
{
	struct sockaddr_in servaddr;
	int ret;

	ret = dccp_get_socket();
	if (ret < 0)
		return ret;
	listen_fd = ret;

	bzero(&servaddr, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port = htons(conf.dccp_port_arg);
	ret = bind(listen_fd, (struct sockaddr *)&servaddr, sizeof(servaddr));
	if (ret < 0)
		return -E_DCCP_BIND;
	ret = dccp_set_socket(listen_fd);
	if (ret < 0)
		return ret;
	ret = listen(listen_fd, 0);
	if (ret < 0)
		return -E_DCCP_LISTEN;
	PARA_DEBUG_LOG("listening on fd %d\n", listen_fd);
	add_close_on_fork_list(listen_fd);
	mark_fd_nonblock(listen_fd);
	return 1;
}

static void dccp_shutdown_client(struct dccp_client *dc)
{
	PARA_DEBUG_LOG("shutting down %s (fd %d)\n", inet_ntoa(dc->addr.sin_addr),
		dc->fd);
	close(dc->fd);
	del_close_on_fork_list(dc->fd);
	list_del(&dc->node);
	free(dc);
}

/** give up if write would block that many times */
#define DCCP_WRITE_RETRIES 100

static int dccp_write(int fd, const char *buf, size_t len)
{
	size_t size, written = 0;
	int ret, retries = 0;
again:
	size = PARA_MIN(1024, len - written);
	ret = write(fd, buf + written, size);
	if (ret < 0) {
		if (errno != EAGAIN || !retries++ > DCCP_WRITE_RETRIES)
			goto err_out;
		PARA_DEBUG_LOG("EAGAIN #%d@%zd/%zd\n", retries, written, len);
		goto again;
	}
	retries = 0;
	written += ret;
	if (written >= len)
		return written;
	ret = write_ok(fd);
	if (ret > 0)
		goto again;
err_out:
	return -E_DCCP_WRITE;
}

static void dccp_send(long unsigned current_chunk,
		__a_unused long unsigned chunks_sent, const char *buf, size_t len)
{
	struct dccp_client *dc, *tmp;
	int ret, header_len;
	char *header_buf;

	if (listen_fd < 0 || !len)
		return;

	list_for_each_entry_safe(dc, tmp, &clients, node) {
		ret = write_ok(dc->fd);
		if (ret < 0) {
			dccp_shutdown_client(dc);
			continue;
		}
		if (!ret)
			continue;
		if (!dc->header_sent && current_chunk) {
			header_buf = vss_get_header(&header_len);
			if (header_buf && header_len > 0) {
				ret = dccp_write(dc->fd, header_buf, header_len);
				if (ret != header_len) {
					int err = errno;
					PARA_ERROR_LOG("header write: %d/%d (%s)\n",
						ret, header_len, ret < 0?
						strerror(err) : "");
					dccp_shutdown_client(dc);
					continue;
				}
				dc->header_sent = 1;
				ret = write_ok(dc->fd);
				if (ret < 0) {
					dccp_shutdown_client(dc);
					continue;
				}
				if (!ret)
					continue;
			}
		}
//		PARA_DEBUG_LOG("writing %d bytes to fd %d\n", len, dc->fd);
		ret = dccp_write(dc->fd, buf, len);
		if (ret != len)
			dccp_shutdown_client(dc);
	}
}

static void dccp_shutdown_clients(void)
{
	struct dccp_client *dc, *tmp;

	list_for_each_entry_safe(dc, tmp, &clients, node)
		dccp_shutdown_client(dc);
}

static char *dccp_info(void)
{
	static char *buf;
	int num_clients = 0;
	struct dccp_client *dc, *tmp;

	free(buf);
	list_for_each_entry_safe(dc, tmp, &clients, node)
		num_clients++;
	buf = make_message("dccp connected clients: %d\n", num_clients);
	return buf;
}

static char *dccp_help(void)
{
	return make_message("no help available\n");
}

/**
 * the init function of the dccp sender
 *
 * \param s pointer to the dccp sender struct
 *
 * It initializes all function pointers of \a s and starts
 * listening on the given port.
 */
void dccp_send_init(struct sender *s)
{
	int ret;

	INIT_LIST_HEAD(&clients);
	s->info = dccp_info;
	s->send = dccp_send;
	s->pre_select = dccp_pre_select;
	s->post_select = dccp_post_select;
	s->shutdown_clients = dccp_shutdown_clients;
	s->help = dccp_help;
	s->client_cmds[SENDER_ON] = NULL;
	s->client_cmds[SENDER_OFF] = NULL;
	s->client_cmds[SENDER_DENY] = NULL;
	s->client_cmds[SENDER_ALLOW] = NULL;
	s->client_cmds[SENDER_ADD] = NULL;
	s->client_cmds[SENDER_DELETE] = NULL;
	self = s;
	ret = dccp_open();
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
		s->status = SENDER_OFF;
	} else
		s->status = SENDER_ON;
}
