/*
 * Copyright (C) 1997-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file para.h global paraslash definitions */

#include "config.h"

#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h> /* time(), localtime() */
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <ctype.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/un.h> /* needed by create_pf_socket */
#include <string.h>
#include "gcc-compat.h"

/** used in various contexts */
#define MAXLINE 255

/** compute the minimum of \a a and \a b */
#define PARA_MIN(a,b) ((a) < (b) ? (a) : (b))
/** compute the maximum of \a a and \a b */
#define PARA_MAX(a,b) ((a) > (b) ? (a) : (b))
/** compute the absolute value of \a a */
#define PARA_ABS(a) ((a) > 0 ? (a) : -(a))

/** debug loglevel, gets really noisy */
#define DEBUG 1
/** still noisy, but won't fill your disk */
#define INFO  2
/** normal, but significant event */
#define NOTICE 3
/** unexpected event that can be handled */
#define WARNING 4
/** unhandled error condition */
#define ERROR 5
/** system might be unreliable */
#define CRIT 6
/** last message before exit */
#define EMERG 7

/** log messages with lower proirity than that will not be compiled in*/
#define COMPILE_TIME_LOGLEVEL 0

#if DEBUG > COMPILE_TIME_LOGLEVEL
#define PARA_DEBUG_LOG(f,...) para_log(DEBUG, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_DEBUG_LOG(...) do {;} while (0)
#endif

#if INFO > COMPILE_TIME_LOGLEVEL
#define PARA_INFO_LOG(f,...) para_log(INFO, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_INFO_LOG(...) do {;} while (0)
#endif

#if NOTICE > COMPILE_TIME_LOGLEVEL
#define PARA_NOTICE_LOG(f,...) para_log(NOTICE, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_NOTICE_LOG(...) do {;} while (0)
#endif

#if WARNING > COMPILE_TIME_LOGLEVEL
#define PARA_WARNING_LOG(f,...) para_log(WARNING, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_WARNING_LOG(...) do {;} while (0)
#endif

#if ERROR > COMPILE_TIME_LOGLEVEL
#define PARA_ERROR_LOG(f,...) para_log(ERROR, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_ERROR_LOG(...) do {;} while (0)
#endif

#if CRIT > COMPILE_TIME_LOGLEVEL
#define PARA_CRIT_LOG(f,...) para_log(CRIT, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_CRIT_LOG(...) do {;} while (0)
#endif

#if EMERG > COMPILE_TIME_LOGLEVEL
#define PARA_EMERG_LOG(f,...) para_log(EMERG, "%s: " f, __FUNCTION__, __VA_ARGS__)
#else
#define PARA_EMERG_LOG(...)
#endif

/**
 * define a standard log function that always writes to stderr
 *
 * \param loglevel_barrier If the loglevel of the current message
 * is less than that, the message is going to be ignored.
 *
 */
#define INIT_STDERR_LOGGING(loglevel_barrier) \
	__printf_2_3 void para_log(int ll, const char* fmt,...) \
	{ \
		va_list argp; \
		if (ll < loglevel_barrier) \
			return; \
		va_start(argp, fmt); \
		vfprintf(stderr, fmt, argp); \
		va_end(argp); \
	}

/** version text used by various commands if -V switch was given */
#define VERSION_TEXT(prefix) "para_" prefix " " PACKAGE_VERSION "\n" \
	"Copyright (C) 2007 Andre Noll\n" \
	"This is free software with ABSOLUTELY NO WARRANTY." \
	" See COPYING for details.\n" \
	"Written by Andre Noll.\n" \
	"Report bugs to <maan@systemlinux.org>.\n"

/** print out \p VERSION_TEXT and exit if version flag was given */
#define HANDLE_VERSION_FLAG(_prefix, _args_info_struct) \
	if (_args_info_struct.version_given) { \
		printf("%s", VERSION_TEXT(_prefix)); \
		exit(EXIT_SUCCESS); \
	}
/** sent by para_server for commands that expect a data file */
#define AWAITING_DATA_MSG "\nAwaiting Data."
/** sent by para_server if authentication was successful */
#define PROCEED_MSG "\nProceed.\n"
/** length of the \p PROCEED_MSG string */
#define PROCEED_MSG_LEN strlen(PROCEED_MSG)
/** sent by para_client to indicate the end of the command line */
#define EOC_MSG "\nEnd of Command."
/** sent by para_client, followed by the decrypted challenge number */
#define CHALLENGE_RESPONSE_MSG "challenge_response:"

/* gui_common */
int para_open_audiod_pipe(char *);
int read_audiod_pipe(int, void (*)(char *));

/* exec */
int para_exec_cmdline_pid(pid_t *pid, const char *cmdline, int *fds);

/* time */
int tv_diff(const struct timeval *b, const struct timeval *a, struct timeval *diff);
long unsigned tv2ms(const struct timeval*);
void d2tv(double, struct timeval*);
void tv_add(const struct timeval*, const struct timeval *, struct timeval *);
void tv_scale(const unsigned long, const struct timeval *, struct timeval *);
void tv_divide(const unsigned long divisor, const struct timeval *tv,
	struct timeval *result);
int tv_convex_combination(const long a, const struct timeval *tv1,
		const long b, const struct timeval *tv2,
		struct timeval *result);
void ms2tv(const long unsigned n, struct timeval *tv);

/* stat */
enum {
	SI_STATUS_BAR,		SI_STATUS,		SI_NUM_PLAYED,
	SI_MTIME,		SI_LENGTH_MIN,		SI_LENGTH_SEC,
	SI_FILE_SIZE,		SI_STATUS_FLAGS,	SI_FORMAT,
	SI_SCORE,		SI_AUDIO_INFO1,		SI_AUDIO_INFO2,
	SI_AUDIO_INFO3,		SI_DBINFO1,		SI_DBINFO2,
	SI_DBINFO3,		SI_DECODER_FLAGS,	SI_AUDIOD_STATUS,
	SI_PLAY_TIME,		SI_UPTIME,		SI_OFFSET,
	SI_LENGTH,		SI_STREAM_START,	SI_CURRENT_TIME,
	SI_AUDIOD_UPTIME,	SI_SELECTOR,		NUM_STAT_ITEMS
};

int stat_item_valid(const char *item);
int stat_line_valid(const char *);
void stat_client_write(char *msg, int itemnum);
int stat_client_add(int fd, long unsigned mask);
unsigned for_each_line(char *, int, void (*)(char *));
#define FOR_EACH_STAT_ITEM(i) for (i = 0; i < NUM_STAT_ITEMS; i++)

__printf_2_3 void para_log(int, const char*, ...);

/* taken from printf man page */
#define PARA_VSPRINTF(fmt, p) \
{ \
	int n, size = 100; \
	p = para_malloc(size); \
	while (1) { \
		va_list ap; \
		/* Try to print in the allocated space. */ \
		va_start(ap, fmt); \
		n = vsnprintf(p, size, fmt, ap); \
		va_end(ap); \
		/* If that worked, return the string. */ \
		if (n > -1 && n < size) \
			break; \
		/* Else try again with more space. */ \
		if (n > -1) /* glibc 2.1 */ \
			size = n + 1; /* precisely what is needed */ \
		else /* glibc 2.0 */ \
			size *= 2; /* twice the old size */ \
		p = para_realloc(p, size); \
	} \
}
